use smart_leds_trait::SmartLedsWrite;

use crate::buffers::Pixbuf;
use crate::render::{HardwarePixel, Output, PixelView, Sample};
use crate::power::brightness_for_mw;
use crate::geometry::*;
use crate::mappings::*;

use core::fmt::Debug;

impl<P: Pixbuf, T: FastWrite> Debug for StrideOutput<P, T> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        f.debug_struct("StrideOutput").finish()
    }
}

pub struct StrideOutput<P: Pixbuf, T: FastWrite> {
    pixbuf: P,
    stride_map: StrideMapping,
    target: T,
    max_mw: u32,
    brightness: u8
}

impl<P: Pixbuf, T: FastWrite> StrideOutput<P, T> {
    pub fn new(pixbuf: P, stride_map: StrideMapping, target: T, max_mw: u32) -> Self {
        assert!(stride_map.pixel_count <= pixbuf.pixel_count(), "map needs {} pixels, I only have PIXEL_NUM={}", stride_map.pixel_count, pixbuf.pixel_count());
        StrideOutput {
            pixbuf,
            stride_map,
            target,
            max_mw,
            brightness: 255
        }
    }
}

impl<P: Pixbuf, T: FastWrite> Sample for StrideOutput<P, T> {
    type Pixel = P::Pixel;
    fn sample(&mut self, rect: &Rectangle<Virtual>) -> impl PixelView<Pixel = Self::Pixel> {
        StrideSampler::new(&mut self.pixbuf,self.stride_map.select(rect))
    }
}

impl<P: Pixbuf<Pixel=T::Color>, T: FastWrite> Output for StrideOutput<P, T> {
    fn blank(&mut self) {
        self.pixbuf.blank();
    }

    fn commit(&mut self) {
        let b = brightness_for_mw(self.pixbuf.as_milliwatts(), self.brightness, self.max_mw);
        if self.target.fast_write(self.pixbuf.iter_with_brightness(b)).is_err() {
            panic!("Could not write frame!");
        };
    }

    fn on_event(&mut self, event: &crate::events::Event) {
        match event {
            crate::events::Event::PropertyChange("output.brightness", Variant::Byte(new_brightness)) => self.brightness = *new_brightness,
            _ => ()
        }
    }
}

pub trait FastWrite: Send {
    type Target: SmartLedsWrite;
    type Color: HardwarePixel;
    type Error;
    fn fast_write<T, I>(&mut self, iterator: T) -> Result<(), Self::Error> where
    T: IntoIterator<Item = I>,
    I: Into<Self::Color>,
    <T as IntoIterator>::IntoIter: Send;
}

#[cfg(feature="rmt")]
pub mod rmt {
    use ws2812_esp32_rmt_driver::driver::color::LedPixelColorGrb24;
    use smart_leds::SmartLedsWrite;
    use rgb::Rgb;
    use ws2812_esp32_rmt_driver::LedPixelEsp32Rmt;

    use super::FastWrite;

    pub type FastWs2812Esp32Rmt<'a> = LedPixelEsp32Rmt<'a, Rgb<u8>, LedPixelColorGrb24>;

    impl FastWrite for FastWs2812Esp32Rmt<'_> {
        type Color = <Self as SmartLedsWrite>::Color;
        type Error = <Self as SmartLedsWrite>::Error;
        type Target = Self;
        fn fast_write<T, I>(&mut self, iterator: T) -> Result<(), Self::Error> where
            T: IntoIterator<Item = I>,
            I: Into<Self::Color>,
            <T as IntoIterator>::IntoIter: Send {
                self.write_nocopy(iterator)
        }
    }
}

#[cfg(feature="spi")]
pub mod spi {
    use smart_leds::SmartLedsWrite;
    use ws2812_spi::prerendered::Ws2812;

    use crate::render::{Display, Surface};
    use crate::platform::smart_leds_lib::SmartLedDisplay;
    use crate::DisplayInit;
    use super::FastWrite;

    use esp_idf_svc::hal::{
        prelude::*,
        gpio::AnyIOPin,
        spi::{
            config::{Config, DriverConfig},
            Dma,
            SpiBusDriver,
            SpiDriver,
        }
    };

    impl<'a> FastWrite for Ws2812<'_, SpiBusDriver<'a, SpiDriver<'a>>> {
        type Color = <Self as SmartLedsWrite>::Color;
        type Error = <Self as SmartLedsWrite>::Error;
        type Target = Self;

        fn fast_write<T, I>(&mut self, iterator: T) -> Result<(), Self::Error> where
            T: IntoIterator<Item = I>,
            I: Into<Self::Color>,
            <T as IntoIterator>::IntoIter: Send {
                let resp = self.write(iterator);
                if let Err(e) = resp {
                    panic!("Could not write SPI frame! {:?}", e)
                } else {
                    resp
                }
        }
    }

    static mut STATIC_BUFFER: [u8; 400 * 12] = [0; 400 * 12];

    pub struct SPIDisplay {}
    impl DisplayInit for SPIDisplay {
        fn new_display<S: Surface>() -> impl Display<S> {
            let peripherals = Peripherals::take().unwrap();

            let driver = SpiDriver::new_without_sclk(
                peripherals.spi2,
                peripherals.pins.gpio5,
                Option::<AnyIOPin>::None,
                &DriverConfig::new().dma(Dma::Auto(4092))
            ).unwrap();

            let cfg = Config::new().baudrate(3.MHz().into());

            let spi = SpiBusDriver::new(driver, &cfg).unwrap();

            const POWER_VOLTS : u32 = 5;
            const POWER_MA : u32 = 500;
            const MAX_POWER_MW : u32 = POWER_VOLTS * POWER_MA;
            unsafe {
                let target = Ws2812::new(spi, &mut STATIC_BUFFER);
                return SmartLedDisplay::<Ws2812<SpiBusDriver<'_, SpiDriver<'_>>>, S, 310>::new(target, MAX_POWER_MW);
            }
        }
    }
}
